/*
    This file is part of the 'yarec' package.
    Copyright (C) 1998  Ralf W. Stephan <stephan@tmt.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    See also http://www.gnu.org
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#include "yarec.h"
#endif

/*#undef HAVE_POSIXTHREADS  /* as long as we can't test em */
#ifdef HAVE_POSIXTHREADS

#include "thread.h"

#define RING_SIZE 24                             /* must be even and > 2 */
#define MAX_DIST (RING_SIZE/3) 
#define NMUTEXES (RING_SIZE)
#define READER_THRNR 0
#define WRITER_THRNR 1

static void *reader_frame (void*), *writer_frame (void*);

static long (*reader_func)(u32*, long), (*writer_func)(u32*, long);
static u32* buf[RING_SIZE];
static int len[RING_SIZE], rslot, wslot, finished, rerror, werror, closed;
static int buf_size;

int DBuf_init (long (*reader)(u32*, long), long (*writer)(u32*, long), int bs)
{
  int i;
  
  buf_size = (int)((16384.0/176400.0) * bs) + 1;
  buf_size &= 0xfffffffe;
  if (buf_size < 1024) buf_size = 1024;
//g_print("samp_fac=%d buf_size=%d\n", bs, buf_size);

  for (i=0; i<RING_SIZE; i++) {
  	len [i] = 0;
  	if ((buf[i] = (u32*)malloc (sizeof(u32)*buf_size)) == 0)
  		return -1;
  }
  
  reader_func = reader;
  writer_func = writer;
  finished = closed = 0;
  rerror = werror = 0;
  wslot = 0;
  
  for (rslot = 0; rslot < MAX_DIST; rslot++) {
  	long l = reader (buf [rslot], buf_size);
  	if (l <= 0) break;
  	len [rslot] = l;
  }
  
  Thread_init_mutexes (NMUTEXES);
  if (Thread_start (writer_frame, WRITER_THRNR))
  	return -1;
  if (Thread_start (reader_frame, READER_THRNR))
  	return -1;
  return 0;
}

int DBuf_rerror()
{
  return rerror;
}

int DBuf_werror()
{
  return werror;
}

void DBuf_close (int flush)
{
  int i;

  if (closed) return;
  finished = 1;
  Thread_join (WRITER_THRNR);
  for (i=0; i<RING_SIZE; i++)
  	free (buf[i]);
  closed = 1;
}

/*---------------------------------------------------------------*/

static void* reader_frame (void *arg)
{
  Thread_blockallsigs();
  Thread_mutex_lock (rslot);
  
  while ((finished + rerror + werror) == 0) {
  	long l;
  	while (((wslot - rslot + RING_SIZE) % RING_SIZE) < MAX_DIST)
  		Thread_yield();
  	l = len [rslot] = reader_func (buf [rslot], buf_size);
  	if (l == 0 || l < buf_size) 
  		finished = 1;
  	else if (l < 0) 
  		rerror = 1;
  	Thread_mutex_unlock (rslot);
  	rslot = (rslot + 1) % RING_SIZE;
  	Thread_mutex_lock (rslot);
  }
  Thread_mutex_unlock (rslot);
  Thread_exit();
}

static void* writer_frame (void *arg)
{
  Thread_blockallsigs();
  Thread_mutex_lock (wslot);
  
  while (!werror && len [wslot] > 0) {
  	while (((rslot - wslot + RING_SIZE) % RING_SIZE) < MAX_DIST) {
  		if (finished) break;
  		Thread_yield();
  	}
  	if (writer_func (buf [wslot], len [wslot]) < 0) 
  		werror = 1;
  	len [wslot] = 0;
  	Thread_mutex_unlock (wslot);
  	wslot = (wslot + 1) % RING_SIZE;
  	Thread_mutex_lock (wslot);
  }
  Thread_mutex_unlock (wslot);
  Thread_exit();
}

/*---------------------------------------------------------------*/

#else /* NOT HAVE_POSIXTHREADS */

/*---------------------------------------------------------------*/
/* Lo, stranger, hide your eyes, step back, and never attempt      */
/* again to read the following, for this is the biggest perversity */
/* ever produced - and your humble author isn't even ashamed!      */

static long (*reader_func)(u32*, long), (*writer_func)(u32*, long);
static int buf_size, len, finished, closed;
static u32 *buf;

int DBuf_init (long (*reader)(u32*, long), long (*writer)(u32*, long), int bs)
{
  bs >>= 13;
  bs <<= 11;
  buf_size = bs;

  buf = (u32*)malloc (sizeof(u32)*bs);
  if (buf==0) return -1;

  reader_func = reader;
  writer_func = writer;  
  finished = closed = 0;
  
  return 0;
}

int DBuf_rerror()
{
  if (finished) return 0;
  len = reader_func (buf, buf_size);
//g_print("r%d ", len);
  return len<0;
}

int DBuf_werror()
{
  if (finished) return 0;
  finished = len < buf_size;
  if (len <= 0) return 0;
  len = writer_func (buf, len);
//g_print("w%d ", len);
  return len<0;
}

void DBuf_close (int flush)
{
  if (closed) return;
  finished = 1;
  free (buf);
  closed = 1;
}

/*---------------------------------------------------------------*/

#endif /* HAVE_POSIXTHREADS */
